import logging
import math
from typing import Any, List, Tuple

import numpy as np
import pandas as pd
from numba import jit

from quaternions.types import POINT, SCALAR, T_PAIR, T_POINT, VECTOR

ZERO_VECTOR = np.zeros(3)

LOGGER = logging.getLogger("quaternions")


@jit
def mul_two_points(p: T_POINT, q: T_POINT) -> T_POINT:
    """ This is the quaternion multproduct of two 4D `T_POINT`s

    Parameters
    ----------
    p
        left (4D) point
    q
        right (4D) point

    Returns
    -------
    prod
        the quaternion product of p and q

    """

    s = p[0] * q[0] - p[1] * q[1] - p[2] * q[2] - p[3] * q[3]
    x = p[0] * q[1] + p[1] * q[0] + p[2] * q[3] - p[3] * q[2]
    y = p[0] * q[2] - p[1] * q[3] + p[2] * q[0] + p[3] * q[1]
    z = p[0] * q[3] + p[1] * q[2] - p[2] * q[1] + p[3] * q[0]

    return np.array([s, x, y, z])


@jit
def rot_matrix_from_point(q: T_POINT) -> np.ndarray:
    """ Yields the rotation matrix for a point representing a unit quaternion

    Parameters
    ----------
    q
        (4D) point

    Returns
    -------
    mat
        the rotation matrix 

    """

    w, x, y, z = q

    row0 = [
        w ** 2 + x ** 2 - y ** 2 - z ** 2,
        2 * (x * y - w * z),
        2 * (x * z + w * y),
    ]
    row1 = [
        2 * (x * y + w * z),
        w ** 2 - x ** 2 + y ** 2 - z ** 2,
        2 * (y * z - w * x),
    ]
    row2 = [
        2 * (x * z - w * y),
        2 * (y * z + w * x),
        w ** 2 - x ** 2 - y ** 2 + z ** 2,
    ]

    return np.array([row0, row1, row2])


def matrix_from_point(p: T_POINT) -> np.ndarray:
    matrix = np.array(
        [[p[0] + p[1] * 1j, -p[2] - p[3] * 1j], [p[2] - p[3] * 1j, p[0] - p[1] * 1j]]
    )
    return matrix


def is_list(o: Any) -> bool:
    return isinstance(o, List)


def is_tuple(o: Any) -> bool:
    return isinstance(o, Tuple)


def _is_2d_array(a: Any) -> bool:
    return isinstance(a, np.ndarray) and a.ndim == 2


def _is_vector_or_point(o: Any, t: Any, dim: int) -> bool:
    if not isinstance(o, t):
        return False
    if isinstance(o, np.ndarray) and o.ndim != 1:
        return False
    if len(o) != dim:
        return False

    return all(is_scalar(e) for e in o)


def is_scalar(o: Any) -> bool:
    return isinstance(o, SCALAR)


def is_vector(o: Any) -> bool:
    return _is_vector_or_point(o, VECTOR, 3)


def is_point(o: Any) -> bool:
    return _is_vector_or_point(o, POINT, 4)


def is_pair(p: Any) -> bool:
    if not (is_tuple(p) and len(p) == 2):
        return False
    s, v = p
    return is_scalar(s) and is_vector(v)


def is_vector_array(a: Any) -> bool:
    if not (_is_2d_array(a) and a.shape[-1] == 3):
        return False
    return all(is_vector(e) for e in a)


def is_point_array(a: Any) -> bool:
    if not (_is_2d_array(a) and a.shape[-1] == 4):
        return False
    return all(is_point(e) for e in a)


def is_vector_list(l: Any) -> bool:
    return is_list(l) and all(is_vector(e) for e in l)


def is_point_list(l: Any) -> bool:
    return is_list(l) and all(is_point(e) for e in l)


def is_pair_list(l: Any) -> bool:
    return is_list(l) and all(is_pair(e) for e in l)


def pair_to_point(pair: T_PAIR) -> T_POINT:
    s, v = pair
    return [s] + list(v)


def euler_arcsin_numerical_stability(rot_mat_elem: float) -> None:
    sign = np.sign(rot_mat_elem)

    if math.isclose(rot_mat_elem, sign, rel_tol=1e-7):
        LOGGER.warning(
            f"setting rotation matrix element "
            "{rot_mat_elem} to {sign} for numerical "
            "stability of arcsin"
        )
        rot_mat_elem = sign

    return rot_mat_elem


def euler_angles_from_rot_matrix(
    *args: float, swap_sign_1: bool = False, swap_sign_2: bool = False
) -> Tuple[float, float, float]:
    """
    From Philip J. Schneider, David H. Eberly:
    Geometric Tools for Computer Graphics.
    2003, Morgan Kaufmann Publishers
    """
    sin_angle2 = euler_arcsin_numerical_stability(args[0])
    angle2 = np.arcsin(sin_angle2)

    if angle2 < np.pi / 2:
        if angle2 > -np.pi / 2:
            angle3 = np.arctan2(args[1], args[2])
            angle1 = np.arctan2(args[3], args[4])
        else:
            # solution not unique
            angle3 = np.arctan2(args[5], args[6])
            if swap_sign_1:
                angle3 = -angle3
            angle1 = 0.0
    else:
        # solution not unique
        angle3 = np.arctan2(args[7], args[8])
        if swap_sign_2:
            angle3 = -angle3
        angle1 = 0.0

    return angle3, angle2, angle1



from typing import List, Tuple, Union

import numpy as np

# types
SCALAR = (int, float, np.number)
VECTOR = (List, np.ndarray)
POINT = (List, np.ndarray)

# type hints
T_SCALAR = Union[int, float, np.number]
T_VECTOR = Union[List[T_SCALAR], np.ndarray]
ALL_VECTOR = Union[T_VECTOR, List[T_VECTOR], np.ndarray]
T_POINT = Union[List[T_SCALAR], np.ndarray]
ALL_POINT = Union[T_POINT, List[T_POINT], np.ndarray]
T_PAIR = Tuple[T_SCALAR, T_VECTOR]
ALL_PAIR = Union[T_PAIR, List[T_PAIR]]



import pathlib
from setuptools import find_packages, setup

# The parent directory
DIR = pathlib.Path(__file__).parent

# The text of the README file
README = (DIR / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="pyquatlib",
    version="1.0.0",
    description='yet another library for quaternions',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/m-bass/pyquatlib",
    author='Marco Bosatta',
    author_email="mb@xp-obj.com",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    packages=find_packages(exclude=("tests", "docs")),
    include_package_data=True,
    install_requires=['numpy', 'pandas', 'numba'],
    python_requires='>=3',
    entry_points={
        "console_scripts": [
            "realpython=reader.__main__:main",
        ]
    },
    test_suite='tests',
)



import numpy as np
import pandas as pd

from quaternions.utils import ZERO_VECTOR


def transform_cartesian_to_spherical(data: pd.DataFrame) -> pd.DataFrame:
    """
    Conversion from cartesian coordinates :math:`(x, y, z)`
    to spherical coordinates :math:`(r, \\theta, \\phi)`\\.

    Obtains the **radius** :math:`r = \\sqrt{x^2 + y^2 + z^2}`\\,
    the **inclination** (polar angle, zenith, geographic latitude) :math:`\\theta` defined by the formula
    :math:`cos(\\theta) = \\frac{z}{r}` and the **azimuth** (geographic longitude) :math:`\\phi`
    defined by the formula :math:`tan(\\phi) = \\frac{y}{x}`\\.

    The convention here is to measure the inclination from north to south
    and the azimuth starting from the X axis, in counterclockwise (
    positive) orientation

    Parameters
    ----------
    data
        a data frame with 3 columns, containing cartesian coordinates

    Returns
    -------
    spherical_df
        a data frame containing radius, inclination and azimuth

    """  # noqa

    r = data.pow(2).sum(1).apply(np.sqrt)

    x, y, z = (data.iloc[:, c] for c in range(3))

    theta = np.arccos(z / r)

    phi = np.arctan2(y, x)

    return pd.DataFrame({"r": r, "theta": theta, "phi": phi,}, index=data.index)


def transform_spherical_to_cartesian(data: pd.DataFrame) -> pd.DataFrame:
    """
    Conversion from spherical coordinates :math:`(r, \\theta, \\phi)`
    to cartesian coordinates :math:`(x, y, z)`\\.

    Obtains the cartesian coordinates :math:`(x, y, z)` by the formulae:

    :math:`\\begin{equation} \\begin{split}
    x & = & r sin(\\theta) cos(\\phi) \\\\
    y & = & r sin(\\theta) sin(\\phi) \\\\
    z & = & r cos(\\theta)
    \\end{split} \\end{equation}`

    See :func:`cartesian_to_spherical` for the angles conventions.

    Parameters
    ----------
    data
        a data frame with 3 columns, containing spherical coordinates

    Returns
    -------
    cartesian_df
        a data frame containing cartesian coordinates

    """  # noqa
    r, theta, phi = (data.iloc[:, c] for c in range(3))

    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)

    return pd.DataFrame({"x": x, "y": y, "z": z,}, index=data.index)


def euler_rotation(axis: str, angle: float) -> np.ndarray:
    """ Return the (extrinsic) Euler rotation matrix about
    the principal `axis` by `angle`.

    Parameters
    ----------
    axis
        One of the principal axes `x`, `y`, `z`.
        In aircraft terminology, the `z` axis is aka `yaw` or `head`, 
        the `y` axis is aka `pitch` and the `x` axis is aka `roll`
    angle
        The angle of rotation

    Returns
    -------
    rot_mat
        The :math:`3 \\times 3` matrix
        representing the rotation
        about the `axis` by `angle`
    
    Raises
    ------
    ValueError
        if `axis` is not one of `x`, `y`, `z`

    """

    axes = list("xyz")

    if axis not in axes:
        raise ValueError(f"{axis} must be one of {axes}")

    if axis == "x":
        return np.array(
            [
                [1, 0, 0],
                [0, np.cos(angle), -np.sin(angle)],
                [0, np.sin(angle), np.cos(angle)],
            ]
        )

    if axis == "y":
        return np.array(
            [
                [np.cos(angle), 0, np.sin(angle)],
                [0, 1, 0],
                [-np.sin(angle), 0, np.cos(angle)],
            ]
        )

    if axis == "z":
        return np.array(
            [
                [np.cos(angle), -np.sin(angle), 0],
                [np.sin(angle), np.cos(angle), 0],
                [0, 0, 1],
            ]
        )


def cross_product_matrix(vector: np.ndarray) -> np.ndarray:
    """
    Get the skew-symmetric matrix for vector cross product operation.
    Derived from natural Lie algebra isomorphism :math:`R^{3} ~ so(3)`\\.

    Parameters
    ----------
    vector
        a 3 dim vector

    Returns
    -------
    Omega
        skew symmetric matrix

    """

    Omega = np.array(
        [
            [0, -vector[2], vector[1]],
            [vector[2], 0, -vector[0]],
            [-vector[1], vector[0], 0],
        ]
    )

    return Omega


def rotation_axis_angle(axis: np.ndarray, phi: np.float) -> np.ndarray:
    """
    Calculate rotation matrix around given axis by given angle,
    in positive (counterclockwise) orientation.
    Also known as **Rodrigues' formula**:

    :math:`\\begin{equation}
    R = exp(\\phi \\Omega) \\\\
    R = id + sin(\\phi)\\Omega + (1 - cos(\\phi))\\Omega^2
    \\end{equation}`

    where :math:`\\Omega` is the skew-symmetric
    matrix in :math:`\\mathfrak{so}(3)` which corresponds to
    :math:`\\frac{\\vec{x}}{ \\|\\vec{x}\\|}`.
    See Also :func:`cross_product_matrix`

    Parameters
    ----------
    axis
        the rotation axis
    phi
        the rotation angle

    Returns
    -------
    R:
        Rotation matrix around axis :math:`\\frac{\\vec{axis}}
        { \\|\\vec{axis}\\|}` by angle :math:`\\phi`

    Raises
    ------
    ArithmeticError
        if the axis vector is ~ zero

    """  # noqa
    if np.allclose(axis, ZERO_VECTOR):
        raise ArithmeticError("Axis vector almost zero")

    axis = axis / np.sqrt(sum(axis ** 2))
    Omega = cross_product_matrix(axis)

    R = np.identity(3) + np.sin(phi) * Omega + (1 - np.cos(phi)) * Omega @ Omega

    return R



"""
This is a custom implementation of quaternions.

Author: Marco Bosatta

-----------------------------------

Data structures that may be used to construct Quaternions
---------------------------------------------------------

SCALAR
VECTOR              list of  ;  np.array of
POINT               list of  ;  np.array of
PAIR                list of


    Type of quaternions, can be :class:`Quaternion` or a list of them

"""

import math
from typing import Any, List, Tuple, Union

import numpy as np

from quaternions.types import ALL_VECTOR, T_PAIR, T_POINT, T_SCALAR, T_VECTOR
from quaternions.utils import (
    ZERO_VECTOR,
    euler_angles_from_rot_matrix,
    euler_arcsin_numerical_stability,
    is_list,
    is_pair,
    is_pair_list,
    is_point,
    is_point_array,
    is_point_list,
    is_scalar,
    is_tuple,
    is_vector,
    is_vector_array,
    is_vector_list,
    matrix_from_point,
    mul_two_points,
    pair_to_point,
    rot_matrix_from_point,
)

T_QUAT = "Quaternion"
ALL_QUAT = Union[T_QUAT, List[T_QUAT]]

T_DQUAT = "DualQuaternion"
ALL_DQUAT = Union[T_DQUAT, List[T_DQUAT]]

# map from numpy ufunc to binary operation
_UFUNC_TO_BINOP = {
    "add": lambda p, q: p + q,
    "subtract": lambda p, q: p - q,
    "multiply": lambda p, q: p * q,
    "true_divide": lambda p, q: p / q,
}

VALID_AXES = {"x", "y", "z"}


class Quaternion(np.lib.mixins.NDArrayOperatorsMixin):
    """
    A quaternion is represented internally by a 4D `point`,
    with scalar part at index 0 and vector part at index 1 to 3:

    :math:`\\begin{equation}
    point = [s, x, y, z]
    \\end{equation}`

    """

    def __init__(self, obj: Any) -> None:
        """ A Quaternion can be constructed:

        * from a scalar

        * from a (3D) vector

        * from a (4D) point

        * from a (scalar, vector) pair

        * from a quaternion

        The internal representation is stored in self.point

        """

        if is_scalar(obj):
            self._point = np.array([obj, 0, 0, 0])
        elif is_vector(obj):
            self._point = np.concatenate([[0], obj])
        elif is_point(obj):
            self._point = np.array(obj)
        elif is_pair(obj):
            self._point = np.array(pair_to_point(obj))
        elif is_quaternion(obj):
            self._point = obj.point
        else:
            raise TypeError(f"Cannot construct a Quaternion from {obj}")

    ########################################
    #    properties
    ########################################

    @property
    def point(self) -> T_POINT:
        return self._point

    @property
    def scalar_part(self) -> T_SCALAR:
        return self.point[0]

    @property
    def vector_part(self) -> T_VECTOR:
        return self.point[1:]

    @property
    def pair(self) -> T_PAIR:
        return self.scalar_part, self.vector_part

    @property
    def complex_matrix(self) -> np.ndarray:
        """ Quaternion representation as a complex matrix
        (in :math:`SU(2)` for unit quaternions):
        :math:`\\begin{equation}
        q = \\begin{bmatrix} s \\ x \\ y \\ z \\end{bmatrix}
        = \\begin{bmatrix} a_{1} \\ a_{2} \\ b_{1} \\ b_{2} \\end{bmatrix}
        \\longmapsto
        \\begin{pmatrix}
        \\alpha & - \\beta \\\\
        \\overline{\\beta} & \\overline{\\alpha}
        \\end{pmatrix}
        \\end{equation}`
        , where :math:`\\alpha = a_{1} + i a_{2}` and :math:`\\beta
        = b_{1} + i b_{2}`

        """
        return matrix_from_point(self.point)

    @property
    def is_scalar(self) -> bool:
        """
        Check if this quaternion has only a scalar part

        Returns
        -------
        is_vector
            True if the vector part vanishes

        """
        return np.allclose(self.vector_part, ZERO_VECTOR)

    @property
    def is_vector(self) -> bool:
        """
        Check if this quaternion has only a vector part

        Returns
        -------
        is_vector
            True if the scalar part vanishes

        """
        return math.isclose(self.scalar_part, 0.0)

    @property
    def is_unit(self) -> bool:
        """
        Check if this is a `unit` quaternion.
        Unit quaternions are isomorphic to the group :math:`SO(3)`
        of 3D rotations.

        Returns
        -------
        is_unit
            True if this is a `unit` quaternion, False otherwise

        """
        return math.isclose(self.squared_norm, 1.0)

    @property
    def is_zero(self) -> bool:
        return self.is_scalar and self.is_vector

    @property
    def conjugate(self) -> T_QUAT:
        """
        The conjugate quaternion

        Returns
        -------
        conj
            The conjugate quaternion

        """
        conj = Quaternion((self.scalar_part, -self.vector_part))
        return conj

    @property
    def inverse(self) -> T_QUAT:
        if self.is_zero:
            raise ArithmeticError("zero is not invertible")

        ip = pair_to_point((self.scalar_part, -self.vector_part))
        ip /= self.squared_norm

        return Quaternion(ip)

    @property
    def norm(self):
        return np.sqrt(self.squared_norm)

    @property
    def squared_norm(self):
        return sum(self.point ** 2)

    ##############################
    # OPERATOR OVERLOADING
    ##############################

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: ({self.pair})\n"

    def __str__(self) -> str:
        return f"({self.pair})"

    def __eq__(self, other: object) -> bool:
        if not is_quaternion(other):
            return False

        return np.allclose(self.point, other.point)

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    def __hash__(self):
        return hash(tuple(self.point))

    ##############################
    # unary ops
    ##############################

    # positive value: +q
    def __pos__(self) -> T_QUAT:
        return self

    # negative value: -q
    def __neg__(self) -> T_QUAT:
        return Quaternion(-self.point)

    # conjugate: ~q
    def __invert__(self) -> T_QUAT:
        return self.conjugate

    # absolute value abs(q)
    def __abs__(self) -> float:
        return self.norm

    ##############################
    # binary ops
    ##############################

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs) -> ALL_QUAT:
        """
        Make sure that we can do binary arithmetic operations
        using compatible data structures as left operand and
        Quaternions as right operand.
        See Also :class:`numpy.lib.mixins.NDArrayOperatorsMixin`
        """

        lop = inputs[0]
        lop = self.to_quaternion(lop)

        binop = _UFUNC_TO_BINOP.get(ufunc.__name__)
        if binop is None:
            raise NotImplementedError(
                f"numpy ufunc {ufunc.__name__} is not " "supported"
            )

        if is_list(lop):
            result = [binop(o, self) for o in lop]
        else:
            result = binop(lop, self)

        return result

    # helper function for multiplication
    def _mult_2_quaternions(self, p: T_QUAT, q: T_QUAT) -> T_QUAT:
        point_prod = mul_two_points(p.point, q.point)

        return Quaternion(point_prod)

    def __add__(self, other: Any) -> ALL_QUAT:
        other = self.to_quaternion(other)

        if is_list(other):
            return [Quaternion(self.point + o.point) for o in other]
        else:
            return Quaternion(self.point + other.point)

    def __iadd__(self, other: Any) -> ALL_QUAT:
        return self + other

    def __radd__(self, other: Any) -> ALL_QUAT:
        other = self.to_quaternion(other)

        if is_list(other):
            return [o + self for o in other]
        else:
            return other + self

    def __sub__(self, other: Any) -> ALL_QUAT:
        other = self.to_quaternion(other)

        if is_list(other):
            return [Quaternion(self.point - o.point) for o in other]
        else:
            return Quaternion(self.point - other.point)

    def __isub__(self, other: Any) -> ALL_QUAT:
        return self - other

    def __rsub__(self, other: Any) -> ALL_QUAT:
        other = self.to_quaternion(other)

        if is_list(other):
            return [o - self for o in other]
        else:
            return other - self

    def __mul__(self, other: Any) -> ALL_QUAT:
        other = self.to_quaternion(other)

        if is_list(other):
            return [self._mult_2_quaternions(self, o) for o in other]
        else:
            return self._mult_2_quaternions(self, other)

    def __imul__(self, other: Any) -> T_QUAT:
        return self * other

    def __rmul__(self, other: Any) -> ALL_QUAT:
        other = self.to_quaternion(other)

        if is_list(other):
            return [o * self for o in other]
        else:
            return other * self

    def __truediv__(self, other: Any) -> ALL_QUAT:
        other = self.to_quaternion(other)

        if is_list(other):
            if any(o.is_zero for o in other):
                raise ArithmeticError("cannot divide by zero")
            return [self._mult_2_quaternions(self, o.inverse) for o in other]
        else:
            if other.is_zero:
                raise ArithmeticError("cannot divide by zero")
            return self._mult_2_quaternions(self, other.inverse)

    def __itruediv__(self, other: Any) -> T_QUAT:
        return self / other

    def __rtruediv__(self, other: Any) -> ALL_QUAT:
        if self.is_zero:
            raise ArithmeticError("cannot divide by zero")

        other = self.to_quaternion(other)

        if is_list(other):
            return [o / self for o in other]
        else:
            return other / self

    ########################################
    #    instance methods
    ########################################

    def group_conjugate(self, argument: Any) -> ALL_QUAT:
        """
        This is the conjugation operation from a group theoretic perspective.

        Parameters
        ----------
        argument
            a (list, array) of quaternions, scalars or vectors

        Returns
        -------
        result
            :math:`q \\cdot v \\cdot q^{-1}` of arguments v

        """
        result = self * argument * self.inverse
        return result

    def euler_angles(
        self, axis3: str, axis2: str, axis1: str
    ) -> Tuple[float, float, float]:
        """
        Return the 3 Euler angles corresponding to the factorisation
        of the rotation represented by this quaternion into
        :math:`R_{axis3} \\circ R_{axis2} \\circ R_{axis1}`

        Parameters
        ----------
        axis3
            The axis of the 3rd Euler rotation
        axis2
            The axis of the second Euler rotation
        axis1
            The axis of the first Euler rotation

        Returns
        -------
        angle3
            The angle of the Euler rotation around axis3
        angle2
            The angle of the Euler rotation around axis2
        angle1
            The angle of the Euler rotation around axis1

        Raises
        ------
        ValueError
            if the sequence of `axis3`, `axis2`, `axis1` 
            is not a permutation of `x`, `y`, `z`
        ArithmeticError
            if this is not a unit quaternion

        """

        axes = [axis3, axis2, axis1]

        if not set(axes) == VALID_AXES:
            raise ValueError(
                f"{axis3}, {axis2}, {axis1} must be a permutation of {VALID_AXES}"
            )

        if not self.is_unit:
            raise ArithmeticError("This is not a unit quaternion")

        rm = self.rot_matrix()

        if axes == ["x", "y", "z"]:
            return euler_angles_from_rot_matrix(
                rm[0, 2],
                -rm[1, 2],
                rm[2, 2],
                -rm[0, 1],
                rm[0, 0],
                rm[1, 0],
                rm[1, 1],
                rm[1, 0],
                rm[1, 1],
                swap_sign_1=True,
            )

        elif axes == ["x", "z", "y"]:
            return euler_angles_from_rot_matrix(
                -rm[0, 1],
                rm[2, 1],
                rm[1, 1],
                rm[0, 2],
                rm[0, 0],
                -rm[2, 0],
                rm[2, 2],
                rm[2, 0],
                rm[2, 2],
            )

        elif axes == ["y", "x", "z"]:
            return euler_angles_from_rot_matrix(
                -rm[1, 2],
                rm[0, 2],
                rm[2, 2],
                rm[1, 0],
                rm[1, 1],
                -rm[0, 1],
                rm[0, 0],
                rm[0, 1],
                rm[0, 0],
            )

        elif axes == ["y", "z", "x"]:
            return euler_angles_from_rot_matrix(
                rm[1, 0],
                -rm[2, 0],
                rm[0, 0],
                -rm[1, 2],
                rm[1, 1],
                rm[2, 1],
                rm[2, 2],
                rm[2, 1],
                rm[2, 2],
                swap_sign_1=True,
            )

        elif axes == ["z", "x", "y"]:
            return euler_angles_from_rot_matrix(
                rm[2, 1],
                -rm[0, 1],
                rm[1, 1],
                -rm[2, 0],
                rm[2, 2],
                rm[0, 2],
                rm[0, 0],
                rm[0, 2],
                rm[0, 0],
                swap_sign_1=True,
            )

        elif axes == ["z", "y", "x"]:
            return euler_angles_from_rot_matrix(
                -rm[2, 0],
                rm[1, 0],
                rm[0, 0],
                rm[2, 1],
                rm[2, 2],
                -rm[0, 1],
                -rm[0, 2],
                rm[0, 1],
                rm[0, 2],
                swap_sign_2=True,
            )

        else:
            # we should never reach this code
            raise NotImplementedError()

        return angle3, angle2, angle1

    def rot_matrix(self) -> np.ndarray:
        """
        For a unit quaternion, return the corresponding rotation matrix.

        Returns
        -------
        rot_matrix
            the rotation matrix

        Raises
        ------
        ArithmeticError
            if this is not a unit quaternion

        """
        if not self.is_unit:
            raise ArithmeticError("This is not a unit quaternion")

        return rot_matrix_from_point(self.point)

    def axis_angle(self) -> Tuple[np.ndarray, float]:
        """
        For a unit quaternion, return the axis and angle of the
        corresponding rotation matrix.

        Returns
        -------
        axis, angle
            the normalized axis and the angle of rotation

        Raises
        ------
        ArithmeticError
            if this is not a unit quaternion

        """
        if not self.is_unit:
            raise ArithmeticError("This is not a unit quaternion")

        if self.is_scalar:
            # represents the identity element: angle is zero, and
            #    we could return any unit axis
            return np.array([1, 0, 0]), 0.0

        angle = 2 * np.arccos(self.scalar_part)
        axis = self.vector_part / np.sqrt(1 - self.scalar_part ** 2)
        return axis, angle

    ########################################
    #    class methods
    ########################################

    @classmethod
    def to_quaternion(cls, obj: Any) -> ALL_QUAT:
        """

        Parameters
        ----------
        obj

        Returns
        -------

        """

        # it's already of type `ALL_QUAT`
        if is_quaternion_type(obj):
            return obj

        # it's a list
        if (
            is_vector_list(obj)
            or is_vector_array(obj)
            or is_point_list(obj)
            or is_point_array(obj)
            or is_pair_list(obj)
        ):
            return [cls(p) for p in obj]

        # one element
        return cls(obj)

    @classmethod
    def rotation_quat_from_axis_angle(cls, axis: T_VECTOR, angle: T_SCALAR) -> T_QUAT:
        """
        Create a rotation quaternion from axis and angle.
        The resulting unit quaternion :math:`r` defines a rotation
        around `axis` by `angle`.
        """

        if not is_vector(axis):
            raise ValueError(f"axis {axis} needs to be a vector")
        if not is_scalar(angle):
            raise ValueError(f"angle {angle} needs to be a scalar")
        if np.allclose(axis, ZERO_VECTOR):
            raise ArithmeticError(f"axis {axis} is almost zero")

        v = axis / np.sqrt(sum(a ** 2 for a in axis))

        s = np.cos(angle / 2)
        u = np.sin(angle / 2) * v

        return Quaternion((s, u))

    @classmethod
    def rotate_by_axis_angle(
        cls, axis: T_VECTOR, angle: T_SCALAR, vectors: ALL_VECTOR
    ) -> ALL_VECTOR:
        """
        Rotate a bunch of vectors by axis, angle
        """

        if not (
            is_vector(vectors) or is_vector_list(vectors) or is_vector_array(vectors)
        ):
            raise ValueError(f"wrong data type for vectors")

        q = cls.rotation_quat_from_axis_angle(axis, angle)
        rotated = q.group_conjugate(vectors)

        if is_list(rotated):
            return [q.vector_part for q in rotated]
        else:
            return rotated.vector_part


def is_quaternion(o: Any) -> bool:
    return isinstance(o, Quaternion)


def is_quaternion_list(l: Any) -> bool:
    return is_list(l) and all(is_quaternion(e) for e in l)


def is_quaternion_type(o: Any) -> bool:
    return is_quaternion(o) or is_quaternion_list(o)


class DualQuaternion(np.lib.mixins.NDArrayOperatorsMixin):
    """
    A dual quaternion is represented internally by two quaternions
    `p` and `q`:

    :math:`\\begin{equation}
    \\sigma = p + \\epsilon q
    \\end{equation}`

    where :math:`\\epsilon` is the ``dual unit`` obeying the rule
    :math:`\\epsilon^2 = 0`.

    """

    def __init__(self, *args: Any) -> None:

        errmsg = f"Cannot construct a dual quaternion from {args}"

        if len(args) == 1:
            (t,) = args
            if not (is_tuple(t) and all(is_quaternion(a) for a in t)):
                raise TypeError(errmsg)
            self._quats = t

        elif len(args) == 2:
            if not all(is_quaternion(a) for a in args):
                raise TypeError(errmsg)
            self._quats = tuple(args)

        else:
            raise TypeError(errmsg)

    ########################################
    #    properties
    ########################################

    @property
    def quats(self) -> Tuple[T_QUAT, T_QUAT]:
        return self._quats

    @property
    def is_zero(self) -> bool:
        return all(x.is_zero for x in self.quats)

    @property
    def conjugate(self) -> T_DQUAT:
        """
        The conjugate dual quaternion
        :math:`\\sigma^* = p^* + \\epsilon q^*`

        Returns
        -------
        conj
            The conjugate dual quaternion

        """
        p, q = self.quats
        return DualQuaternion(p.conjugate, q.conjugate)

    @property
    def inverse(self) -> T_DQUAT:
        """ The inverse of :math:`\\sigma = p + \\epsilon q`
        exists if :math:`p` is invertible and
        is :math:`\\sigma^{-1} = p^{-1} - \\epsilon p^{-1} q p^{-1}`
        """

        p, q = self.quats

        if p.is_zero:
            raise ArithmeticError(f"{self} is not invertible")

        ip = p.inverse
        iq = -ip * q * ip

        return DualQuaternion(ip, iq)

    @property
    def double_conjugate(self) -> T_DQUAT:
        """
        The double conjugate dual quaternion
        :math:`\\sigma^* = p^* - \\epsilon q^*`

        Returns
        -------
        conj
            The double conjugate dual quaternion

        """
        p, q = self.quats
        return DualQuaternion(p.conjugate, -q.conjugate)

    @property
    def is_unit(self) -> bool:
        norm = self * self.conjugate
        p, q = norm.quats
        return p == Quaternion(1) and q.is_zero

    ##############################
    # OPERATOR OVERLOADING
    ##############################

    def __repr__(self) -> str:
        p, q = self.quats
        return f"{self.__class__.__name__}: ({p}; {q})\n"

    def __str__(self) -> str:
        p, q = self.quats
        return f"({p}; {q})"

    def __eq__(self, other: object) -> bool:
        if not is_dual_quaternion(other):
            return False

        return all(x == y for x, y in zip(self.quats, other.quats))

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    def __hash__(self):
        return hash(self.quats)

    ##############################
    # binary ops
    ##############################

    # helper function for multiplication
    def _mult_2_dual_quaternions(self, a: T_DQUAT, b: T_DQUAT) -> T_DQUAT:

        p, q = a.quats
        po, qo = b.quats
        return DualQuaternion(p * po, p * qo + q * po)

    def __add__(self, other: ALL_DQUAT) -> ALL_DQUAT:

        if is_list(other):
            return [
                DualQuaternion(tuple(x + y for x, y in zip(self.quats, o.quats)))
                for o in other
            ]
        else:
            return DualQuaternion(tuple(x + y for x, y in zip(self.quats, other.quats)))

    def __sub__(self, other: ALL_DQUAT) -> ALL_DQUAT:

        if is_list(other):
            return [
                DualQuaternion(tuple(x - y for x, y in zip(self.quats, o.quats)))
                for o in other
            ]
        else:
            return DualQuaternion(tuple(x - y for x, y in zip(self.quats, other.quats)))

    def __mul__(self, other: ALL_DQUAT) -> ALL_DQUAT:

        if is_list(other):
            return [self._mult_2_dual_quaternions(self, o) for o in other]
        else:
            return self._mult_2_dual_quaternions(self, other)

    def __truediv__(self, other: ALL_DQUAT) -> ALL_DQUAT:

        if is_list(other):
            if any(o.is_zero for o in other):
                raise ArithmeticError("cannot divide by zero")
            return [self._mult_2_dual_quaternions(self, o.inverse) for o in other]
        else:
            if other.is_zero:
                raise ArithmeticError("cannot divide by zero")
            return self._mult_2_dual_quaternions(self, other.inverse)

    ########################################
    #    class methods
    ########################################

    @classmethod
    def rigid_displacement(
        cls, axis: T_VECTOR, angle: T_SCALAR, translation: T_VECTOR, rotate_first=True
    ) -> T_DQUAT:
        """
        Create a unit dual quaternion representing a rigid displacement:
        rotation around `axis` by `angle` and `translation`.
        If `rotate_first`, rotation happens before translation, otherwise,
        translation before rotation.
        """

        r = Quaternion.rotation_quat_from_axis_angle(axis, angle)

        if not is_vector(translation):
            raise ValueError(f"translation {translation} needs to be a vector")

        t = np.array(translation) / 2
        if rotate_first:
            q = t * r
        else:
            q = r * t

        return DualQuaternion(r, q)


def is_dual_quaternion(o: Any) -> bool:
    return isinstance(o, DualQuaternion)


def is_dual_quaternion_list(l: Any) -> bool:
    return is_list(l) and all(is_dual_quaternion(e) for e in l)


def is_dual_quaternion_type(o: Any) -> bool:
    return is_dual_quaternion(o) or is_dual_quaternion_list(o)
